@@ -5,6 +5,67 @@ import re
 import formatter
 import util
 
+from HTMLParser import HTMLParser
+from htmlentitydefs import entitydefs
+
+class JavadocHtmlParser(HTMLParser):
+    def __init__(self):
+        super(JavadocHtmlParser, self).__init__(self)
+
+        self.pending_tag = None
+        self.pending_template = None
+        self.pending_data = []
+        self.output = []
+
+    def build(self):
+        self.close()
+        return ''.join(self.output)
+
+    def handle_starttag(self, tag, attrs):
+        if self.pending_tag:
+            return
+
+        if tag == 'p':
+            self.output.append('\n\n')
+            return
+
+        attrs = dict(attrs)
+        self.pending_tag = tag
+        self.pending_data = []
+
+        if tag == 'a' and 'href' in attrs:
+            self.pending_template = '`%%s <%s>`_' % (attrs['href'],)
+        elif tag == 'b':
+            self.pending_template = '**%s**'
+        elif tag == 'i':
+            self.pending_template = '*%s*'
+        elif tag == 'tt':
+            self.pending_template = '``%s``'
+        else:
+            self.pending_tag = None
+
+    def handle_endtag(self, tag):
+        if tag == self.pending_tag:
+            self.output.append(self.pending_template % (''.join(self.pending_data)))
+            self.pending_tag = None
+
+    def handle_data(self, data):
+        if self.pending_tag:
+            self.pending_data.append(data)
+        else:
+            self.output.append(data)
+
+    def handle_entityref(self, name):
+        self.handle_data(entitydefs[name])
+
+    def handle_charref(self, name):
+        if name[0] == 'x':
+            value = int(name[1:], 16)
+        else:
+            value = int(name)
+
+        self.handle_data(unichr(value))
+
 class JavadocRestCompiler(object):
     """ Javadoc to ReST compiler. Builds ReST documentation from a Java syntax
     tree. """
@@ -34,16 +95,24 @@ class JavadocRestCompiler(object):
         # Format all bolds
         s = re.sub(r'<b>(.*?)</b>', r'**\1**', s)
 
-        # Reformat all links
+        # Format all code
+        s = re.sub(r'<tt>(.*?)</tt>', r'``\1``', s)
+
+        # Reformat all internal links
         s = re.sub(r'\{@link\s+([^\s}]+)\s*\}',
                    lambda m: ':java:ref:`%s`' % (m.group(1).replace('#', '.'),),
                    s)
 
-        # Reformat all links with labsl
+        # Reformat all internal links with labsl
         s = re.sub(r'\{@link\s+([^\s}]+)\s+([^\s}]+)\s*\}',
                    lambda m: ':java:ref:`%s %s`' % (m.group(2), m.group(1).replace('#', '.')),
                    s)
 
+        # Reformat all HTML links
+        s = re.sub(r'''<a href=(.*?)>(.*?)</a>''',
+                   lambda m: '`%s <%s>`_' % (m.group(2).strip(), m.group(1).strip('"').strip("'")),
+                   s)
+
         def sub_list_items(list_type, ul):
             return '\n\n%s\n' % (re.sub(r'<li>\s*(.*?)\s*</li>\s*', r'%s \1\n' % (list_type,), ul.strip()),)
 
@@ -75,7 +144,7 @@ class JavadocRestCompiler(object):
         if doc.author:
             output.add_line(':author: %s' % (self.__html_to_rst(doc.author),))
 
-        for name, value in doc.params.items():
+        for name, value in doc.params:
             output.add_line(':param %s: %s' % (name, self.__html_to_rst(value)))
 
         if doc.return_doc:
